from sympy import Array, Basic, Expr, Mul, S, Symbol, diff
from sympy.core.decorators import call_highest_priority
from sympy.tensor.tensor import Tensor as SympyTensor

from .tensor import Tensor


class DiffOperator(Expr):
    _op_priority = 11
    is_commutative = False

    def __new__(cls, *args, left=None):
        if left == S.Zero:
            return S.Zero
        obj = Basic.__new__(cls, *args)
        obj.left = left if left is not None else S.One
        return obj

    def __repr__(self):
        syms = [s.name for s in self.args]
        name = '\u2202(%s)' % ','.join(syms)
        return str(Mul(self.left, Symbol(name, commutative=False)))

    def __str__(self):
        return self.__repr__()

    @call_highest_priority('__rmul__')
    def __mul__(self, other):
        if isinstance(other, DiffOperator):
            args = self.args + other.args
            return DiffOperator(*args)
        return Mul(self.left, diff(other, *self.args)).doit()

    @call_highest_priority('__mul__')
    def __rmul__(self, other):
        return DiffOperator(*self.args, left=Mul(other, self.left))


class PartialDerivative(Tensor):
    is_TensorDerivative = True

    def __new__(cls, metric, **kwargs):
        basis = list(map(DiffOperator, metric.coords))
        return super().__new__(
            cls, '\u2202', basis, metric, comm='partial', covar=(-1,)
        )

    def __repr__(self):
        return self._print()

    def __str__(self):
        return self.__repr__()



"""
Name: Bianchi IV
Coordinates: Cartesian
"""
from sympy import Function, diag, exp, symbols

coords = symbols('t x y z', real=True)
variables = ()
functions = symbols('alpha beta gamma', cls=Function)
t, x, y, z = coords
al, be, ga = functions
metric = diag(1, exp(al(t)), exp(be(t)), exp(ga(t)))



"""
Name: Generic Static Spherical
References: Stephani (14.1) p163
Coordinates: Spherical
Symmetry:
    - Spherical
    - Static
"""
from sympy import Function, diag, exp, sin, symbols

coords = symbols('t r theta phi', real=True)
variables = ()
functions = symbols('alpha beta', cls=Function)
t, r, th, ph = coords
al, be = functions
metric = diag(-exp(2 * al(r)), exp(2 * be(r)), r ** 2, r ** 2 * sin(th) ** 2)
